/*****************************************************************
 * Copyright (c) ParaSoft Corporation 1991.  All rights reserved.*
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft       *
 * The copyright notice above does not evidence any              *
 * actual or intended publication of such source code.           *
 *****************************************************************/

#include "extern.h"

int over_in = 0;

/**/U_DEFINE_STACK (overs, T_Entry*, 10)

static char *overname;

over_begin (spec)
Node *spec;
{
	T_Entry *entry;

	if (spec == NN) {
		overname = (char*) 0;
		return;
	}

	overname = spec->GenericSpec.Name->com.str;

	entry = install_glo (overname, TAB_NAME, 0);
	SET_over (entry);
	
	over_in = 1;
}

over_end ()
{
	int i;
	T_Entry *entry;
	Over *over;

	if (overname == (char*) 0)
		return;

	over = (Over*) U_malloc (sizeof (Over));
	over->name = overname;
	over->nover = overs_top;
	over->entry = (T_Entry**) U_malloc (sizeof (T_Entry*) * overs_top);

	for (i = 0; i < overs_top; i++)
		over->entry [i] = overs [i];

	entry = lookup_user (overname);

	ENTRY_OVER (entry) = over;
	
	over_in = 0;
	U_stack_clear_overs ();
}

#ifdef CHECK
over_show (over)
Over *over;
{
	int i;
	T_Entry *entry;

	if (over == (Over*) 0)
		return;

	printf (" Over:\n");

	for (i = 0; i < over->nover; i++) {
		putchar ('\t');
		table_show_entry (over->entry [i]);
		if (i != over->nover - 1)
			putchar ('\n');
	}
}
#endif

T_Entry *over_replace (entry, sublist)
T_Entry *entry;
Node *sublist;
{
	Over *over = ENTRY_OVER (entry);
	int i, nover = over->nover, num, nargs;
	Interf *interf;
	Node *sub, *temp;

	for (i = 0; i < nover; i++) {
		interf = over->entry[i]->interf;
		nargs = interf->nargs;

		num = 1;
		for (LIST_LOOP (sub, sublist, temp)) {
			if (sub != NN
			&& !over_typ_eq (sub->Subscript.Expr->com.typ, interf->entry[num]->typ))
				goto nextover;

			num++;
		}

		if (nargs > num && !IS_optional (interf->entry[num]))
			goto nextover;

		return over->entry[i];

nextover:
		;
	}

	if (!IS_intrinsic (entry) && !IS_compile_time (entry))
		goto no_match;

	if (IS_elemental (entry))
		goto for_elemental;

/*!!! tries again with only if it is array or not */

	for (i = 0; i < nover; i++) {
		interf = over->entry[i]->interf;
		nargs = interf->nargs;

		num = 1;
		for (LIST_LOOP (sub, sublist, temp)) {
			if (sub != NN
			&& !over_shape_eq (sub->Subscript.Expr->com.typ
			            , interf->entry[num]->typ))
				goto nextover2;

			num++;
		}

		if (nargs > num && !IS_optional (interf->entry[num]))
			goto nextover2;

		return over->entry[i];

nextover2:
		;
	}

for_elemental:
		
/*!!! tries again with only the types */

	for (i = 0; i < nover; i++) {
		interf = over->entry[i]->interf;
		nargs = interf->nargs;

		num = 1;
		for (LIST_LOOP (sub, sublist, temp)) {
			if (sub != NN
			&& etype_ne (sub->Subscript.Expr->com.typ->etype
			            , interf->entry[num]->typ->etype))
				goto nextover1;

			num++;
		}

		if (nargs > num && !IS_optional (interf->entry[num]))
			goto nextover1;

		return over->entry[i];

nextover1:
		;
	}

no_match:

	sprintf (serr, "no match for overloaded procedure '%s'", entry->orig_name);
	fatal (serr);

	return entry;
}

static over_typ_eq (t1, t2)
Typ *t1, *t2;
{
	int ret = 1;
	Shape *s1 = t1->shape, *s2 = t2->shape;

	if (etype_ne (t1->etype, t2->etype))
		ret = 0;

	if (s1 != SN || s2 != SN) {
		if (s1 == SN)
			ret = 0;
		else if (s2 == SN)
			ret = 0;
		else
			if (s1->rank != s2->rank)
				ret = 0;
	}

	return ret;
}

static over_shape_eq (t1, t2)
Typ *t1, *t2;
{
	int ret = 1;
	Shape *s1 = t1->shape, *s2 = t2->shape;

	if (s1 != SN || s2 != SN) {
		if (s1 == SN)
			ret = 0;
		else if (s2 == SN)
			ret = 0;
		else
			ret = 1;
	}

	return ret;
}
